/*
 * Copyright (C) 2005 Bernhard Erb

 *This program is free software; you can redistribute it and/or
 *modify it under the terms of the GNU General Public License
 *as published by the Free Software Foundation; either version 2
 *of the License, or (at your option) any later version.

 *This program is distributed in the hope that it will be useful,
 *but WITHOUT ANY WARRANTY; without even the implied warranty of
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *GNU General Public License for more details.

 *You should have received a copy of the GNU General Public License
 *along with this program; if not, write to the Free Software
 *Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/************************************************************
 * project= tweety                                          *
 * codename= kleiner eibnet/ip server                       *
 * version= 1.1                                             *
 ************************************************************/

/**
 * @file receive_udp.c
 *
 * This file contains the code for the UDP receveive process.
 *
 * @date created 16.05.05,
 *       last modified 02.09.05
 *
 * @author bernhard erb
 *         bernhard.erb@gmx.net
 *
 */


/*
 * ********************************************* includes ***
 */

#include "common.h"

/*
 * ********************************************** globals ***
 */

int send_eib_pipe;

/*
 * ******************************************* prototypes ***
 */

void sendPacket(EIBNETIP_PACKET * packet, struct sockaddr_in address);	/* forward declaration: method in header.c */
void bailOut(char *message);	/* forward declaration: method in header.c */

/*
 * ******************************************** functions ***
 */

/**
 * This method constructs a server HPAI 
 * at the specified location 
 */

void makeServerHpai(byte * address, struct sockaddr_in clientaddress)
{
#ifndef SECURE
    FILE *fp;
    char pre[160];
    char post[120];
    char src[160];
#endif

    /*
     * *** get local ip address ***
     */
    EIBNETIP_HPAI *hpai = (EIBNETIP_HPAI *) address;
    hpai->structlength = sizeof(EIBNETIP_HPAI);
    hpai->hostprotocol = IPV4_UDP;
    hpai->port = htons(SERVERPORT);

    /*
     * if server is in secure mode the serveraddress is taken
     * from define in header.h. Else the address is extracted
     * using /sbin/ip route get... 
     */

#ifndef SECURE


    strcpy(pre, "/sbin/ip route get ");
    strcpy(src, inet_ntoa(clientaddress.sin_addr));
    strcpy(post,
	   " | awk '/src/ {split($0,bla,\"src \");printf bla[2]}' > /tmp/help");
    strncat(pre, src, 160);
    strncat(pre, post, 160);

    system(pre);		/* execute command */

    if ((fp = fopen("/tmp/help", "r")) == NULL) {	/* open result file */
	bailOut("cant create file: use secure mode ");
    }

    fgets(pre, 32, fp);		/* read result from file */
    unlink("/tmp/help");
    *((unsigned long *) &(hpai->ip1)) = inet_addr(pre);
#else
    *((unsigned long *) &(hpai->ip1)) = inet_addr(SERVERADDRESS);
#endif
}

/**
 * This method is usefull to construct
 * a device information descriptor 
 * at the specified location 
 */

void make_DEV_INF_DIB(byte * offset)
{
    EIBNETIP_DEVINF_DIB *devinf = (EIBNETIP_DEVINF_DIB *) offset;
    devinf->structlength = sizeof(EIBNETIP_DEVINF_DIB);
    devinf->descriptiontypecode = DEVICE_INFO;
    devinf->knxmedium = TP1;
    devinf->devicestatus = 0x00;	/* If CPU is in program mode 0x01 */
    devinf->eibaddress = address;	/* get the Adress of the EIB device */
    devinf->projectinstallationidentifier = 0x0000;	/* project identifier */
    devinf->serialnumber[0] = 0x00;	/* we have no serialnumber */
    devinf->serialnumber[1] = 0x00;
    devinf->serialnumber[2] = 0x00;
    devinf->serialnumber[3] = 0x00;
    devinf->serialnumber[4] = 0x00;
    devinf->serialnumber[5] = 0x00;
    *((unsigned long *) (devinf->multicastaddress)) =
	inet_addr(MULTICASTADDRESS);
    devinf->macaddress[0] = 0x00;	/* EIBNETIP device MAC address who knows it? */
    devinf->macaddress[1] = 0x00;
    devinf->macaddress[2] = 0x00;
    devinf->macaddress[3] = 0x00;
    devinf->macaddress[4] = 0x00;
    devinf->macaddress[5] = 0x00;
    memcpy(devinf->name, friendlyName, 24);
}

/**
 * This method constructs a new 
 * supported families description 
 * at the specified address 
 */
void make_SUPPFAMDIB(byte * address)
{
    EIBNETIP_SUPPFAM_DIB *suppfam = (EIBNETIP_SUPPFAM_DIB *) (address);
    suppfam->structlength = sizeof(EIBNETIP_SUPPFAM_DIB);	/*length */
    suppfam->descriptiontypecode = SUPP_SVC_FAMILIES;
    ((byte *) suppfam)[2] = EIBNETIP_CORE;
    ((byte *) suppfam)[3] = 0x01;
    ((byte *) suppfam)[4] = EIBNETIP_TUNNELING;
    ((byte *) suppfam)[5] = 0x01;
}

/** 
 * This method constructs a sockaddress object
 * out of a EIBNETIP_HPAI. Usefull for answering 
 */
struct sockaddr_in make_Sockaddr(byte * eibnetip_hpai)
{
    struct sockaddr_in addr;
    EIBNETIP_HPAI *hpai = (EIBNETIP_HPAI *) eibnetip_hpai;
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = *((unsigned long *) &hpai->ip1);
    addr.sin_port = hpai->port;
    return addr;
}

/**
 * Sends a disconnect request and 
 * closes the connection by setting
 * needed flags in the shared memory 
 */

void sendDisconnect_Request()
{
    EIBNETIP_DISCONNECT_REQUEST *discon;
    /*
     * *** prepare packet ***
     */
    if (active_connection->reset == 0) {
	EIBNETIP_PACKET *answer = (EIBNETIP_PACKET *) malloc(HEADER_SIZE_10 + sizeof(EIBNETIP_DISCONNECT_REQUEST));	/*send disconnect request */
	answer->head.headersize = HEADER_SIZE_10;
	answer->head.version = EIBNETIP_VERSION_10;
	answer->head.servicetype = htons(DISCONNECT_REQUEST);
	answer->head.totalsize =
	    htons(HEADER_SIZE_10 + sizeof(EIBNETIP_DISCONNECT_REQUEST));
	discon =
	    (EIBNETIP_DISCONNECT_REQUEST *) ((byte *) answer +
					     HEADER_SIZE_10);

	discon->channelid = active_connection->channelid;

	sem_wait(semid);	/* is shared memory available ? */
	makeServerHpai((byte *) & (discon->controlendpoint),
		       active_connection->control_endpoint);


	sendPacket(answer, active_connection->control_endpoint);


	active_connection->reset = 1;
	sem_post(semid);	/* free shared memory */



	free(answer);
    }
}

/**
 * This method acts as signal handler
 * for the receive udp subprocess.
 * Captured signals:
 * - SIGALRM: Timout signal for client
 * heartbeating
 */
void ReceiveUdp_HandleSignal(int sig)
{
    switch (sig) {
    case SIGALRM:		/* heartbeat time out reached */
	{
#ifdef DEBUG
	    fprintf(stdout, "DEBUG: timeout reached user disconnected\n");
	    sendDisconnect_Request();
#endif
	}
	break;
    }
}

/**
 * Inits the receive udp subprocess:
 * - opens shared memory
 * - inits semaphor
 * - inits receive pipe
 * - registers signals
 */
void init_receiveUdp()
{
    /*
     * *** open shared memory ***
     */
    if ((active_connection =
	 (connection *) shmat(nSharedMem, (const void *) 0,
			      0)) == (connection *) - 1) {
	bailOut("Error: can't attach memory");
    }
    active_connection->reset = 1;	/* init connection object */
    active_connection->channelid = -1;

    /*
     * ***************************** init semaphor ***
     */

    if ((semid = sem_open(KEY_SEMAPHOR, 0, PERM, 2)) == (sem_t *) NULL) {
	bailOut("Error: can't grab semaphor");
    }

    /*
     * **************************** init receive pipe ***
     */

    if ((send_eib_pipe = open(RECEIVE_PIPE, O_WRONLY)) == -1) {
	bailOut("Error: can't open pipe receive udp");
    }

    /*
     * ****************************** register signal ***
     */

    signal(SIGALRM, ReceiveUdp_HandleSignal);
}


/**
 * Main method for receive udp 
 * subprocess. First calls init_receiveUDP
 * method, then blocks for arriving UDP packets.
 * Depending on the packet type, different 
 * actions are taken. Search and description
 * request are immediately answered by this
 * process.
 * cEMI packets from tunnelling requests are 
 * send without alteration to the receive pipe.
 * If a Tunnelling request is received the 
 * SIGUSR1 signal is sent to the sendUDP 
 * subprocess. 
 */
void receiveUdp()
{
    EIBNETIP_COMMON_CONNECTION_HEADER *head;
    EIBNETIP_DISCONNECT_RESPONSE *disconnect_response;
    EIBNETIP_CONNECTIONSTATE_RESPONSE *connectionstate_response;
    struct sockaddr_in addr;
    byte received[1024];
    int length = sizeof(addr);
    byte channelid = 100;
    EIBNETIP_PACKET *received_packet;

    init_receiveUdp();		/* allocate ressources */


    while (1) {
	/* listen for arriving udp packets */

	if (recvfrom
	    (sock, received, 1024, 0, (struct sockaddr *) &addr,
	     (socklen_t *) & length) < 0) {
	    bailOut("Error: can't receive packet: probably socket closed");
	}
	received_packet = (EIBNETIP_PACKET *) received;

	/* witch packet is it? */

	switch (htons(received_packet->head.servicetype)) {

	case (DESCRIPTION_REQUEST):
	    {

#ifdef DEBUG
		fprintf(stdout, "DEBUG: Description request received \n");
		zeig(received,htons(received_packet->head.totalsize));
#endif

		/* if it has not the requested length discard it */
		if (htons(received_packet->head.totalsize) ==
		    HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI)) {

		    EIBNETIP_HPAI *hpai =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10);

		    /*
		     * *** prepare description response ***
		     */
		    EIBNETIP_PACKET *answer = malloc(HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI) + sizeof(EIBNETIP_DEVINF_DIB) + 6);	/*6 is sizeof eibnetip_suppfam_dib */

		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype = htons(DESCRIPTION_RESPONSE);

		    answer->head.totalsize =
			htons(HEADER_SIZE_10 +
			      sizeof(EIBNETIP_DEVINF_DIB) +
			      sizeof(EIBNETIP_SUPPFAM_DIB));

		    /* attach device information */
		    make_DEV_INF_DIB((byte *) answer + HEADER_SIZE_10);
		    /* attach supported families inofrmation */
		    make_SUPPFAMDIB((byte *) answer + HEADER_SIZE_10 +
				    sizeof(EIBNETIP_DEVINF_DIB));

		    /* send packet */
		    sendPacket(answer, make_Sockaddr((byte *) hpai));
		}
	    }
	case (CONNECT_REQUEST):
	    {

#ifdef DEBUG
		fprintf(stdout, "DEBUG: Connection request received \n");
		zeig(received,htons(received_packet->head.totalsize));
#endif

		/* if it has not the descripted lenth discard it */
		if (htons(received_packet->head.totalsize) ==
		    HEADER_SIZE_10 + 2 * sizeof(EIBNETIP_HPAI) +
		    sizeof(EIBNETIP_CRI_CRD)) {

		    /*
		     * *** extract connect request ***
		     */
		    EIBNETIP_HPAI *control_endpoint =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10);

		    EIBNETIP_HPAI *data_endpoint =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10 +
					   sizeof(EIBNETIP_HPAI));

		    EIBNETIP_CRI_CRD *cri =
			(EIBNETIP_CRI_CRD *) ((byte *) received_packet +
					      HEADER_SIZE_10 +
					      2 * sizeof(EIBNETIP_HPAI));

		    /*
		     * *** prepare connect response ***
		     */
		    EIBNETIP_PACKET *answer =
			malloc(HEADER_SIZE_10 + 2 + sizeof(EIBNETIP_HPAI) +
			       sizeof(EIBNETIP_CRI_CRD));
		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype = htons(CONNECT_RESPONSE);
		    answer->head.totalsize =
			htons(2 + HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI) +
			      sizeof(EIBNETIP_CRI_CRD));

		    /* attach channelid and status */
		    makeServerHpai((byte *) answer + HEADER_SIZE_10 + 2,
				   addr);

		    /* we accept only tunnel connections */
		    if (cri->connectiontypecode != TUNNEL_CONNECTION) {
#ifdef DEBUG
			fprintf(stdout, "DEBUG: Wrong connection \n");
#endif
			*((byte *) answer + HEADER_SIZE_10 + 1) =
			    E_CONNECTION_TYPE;
		    } else {
			switch (cri->data1) {
			case (TUNNEL_LINKLAYER):	/*we accept only link layer */
			    {
				/* no active connection at the moment */

				if (active_connection->reset) {
				    EIBNETIP_CRI_CRD *crd;
				    /*
				     * *** prepare active connection object ***
				     */
				    channelid++;	/* Make a new connection */
				    sem_wait(semid);	/* shared memory free ? */

				    active_connection->reset = 0;
				    active_connection->channelid =
					channelid;
				    active_connection->lastaccess =
					time((time_t *) NULL);
				    active_connection->control_endpoint =
					make_Sockaddr((byte *)
						      control_endpoint);
				    active_connection->data_endpoint =
					make_Sockaddr((byte *)
						      data_endpoint);
				    active_connection->sequenceSend = 0;
				    active_connection->sequenceReceive = 0;

				    sem_post(semid);	/* freee shared memory */

				    alarm(HEARTBEAT_REQUEST_TIME_OUT);	/* set the timeout */

				    *((byte *) answer + HEADER_SIZE_10) =
					channelid;
				    *((byte *) answer + HEADER_SIZE_10 +
				      1) = 0x00;
				    crd = (EIBNETIP_CRI_CRD *) ((byte *)
								answer +
								HEADER_SIZE_10
								+ 2 +
								sizeof
								(EIBNETIP_HPAI));
				    /* attach the connection information */
				    crd->structlength = 0x04;
				    crd->connectiontypecode =
					EIBNETIP_TUNNELING;
				    crd->data1 = address & 0x00FF;
				    crd->data2 = address >> 8;


#ifdef DEBUG
				    fprintf(stdout,
					    "DEBUG: Connection established \n");
#endif
				} else {	/* we can't accept any other connection */

#ifdef DEBUG
				    fprintf(stdout,
					    "DEBUG: Max connections reached\n");
#endif
				    *((byte *) answer + HEADER_SIZE_10 +
				      1) = E_NO_MORE_CONNECTIONS;
				}
				break;
			    }
			default:	/* this connection type is not supported */
			    {
				EIBNETIP_CRI_CRD *crd;
				/* set error code */
				*((byte *) answer + HEADER_SIZE_10 + 1) =
				    E_CONNECTION_OPTION;
				crd =
				    (EIBNETIP_CRI_CRD *) ((byte *) answer +
							  HEADER_SIZE_10 +
							  2 +
							  sizeof
							  (EIBNETIP_HPAI));
				crd->structlength = 0x04;
				crd->connectiontypecode =
				    E_TUNNELING_LAYER;
				crd->data1 = address & 0x00FF;
				crd->data2 = address >> 8;
#ifdef DEBUG
				fprintf(stdout,
					"DEBUG: connectiontype not supported %X\n",
					cri->data1);
#endif
				break;
			    }

			}

		    }

		    /* send answer to the client */
		    sendPacket(answer,
			       make_Sockaddr((byte *) control_endpoint));

		}

	    }
	case (CONNECTIONSTATE_REQUEST):
	    {
#ifdef DEBUG
	      fprintf(stdout,"DEBUG: Connectionstate request received \n");              zeig(received,htons(received_packet->head.totalsize));
#endif

		/* if the request has not the right size discard it */
		if (htons(received_packet->head.totalsize) ==
		    sizeof(EIBNETIP_HPAI) + 2 + HEADER_SIZE_10) {
		    /*
		     * *** extract connectionstate request ***
		     */
		    EIBNETIP_HPAI *control_endpoint =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10 + 2);
		    byte received_channelid =
			*((byte *) received_packet + HEADER_SIZE_10);

		    /*
		     * *** prepare answer ***
		     */

		    EIBNETIP_PACKET *answer =
			malloc(HEADER_SIZE_10 +
			       sizeof(EIBNETIP_CONNECTIONSTATE_RESPONSE));
		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype =
			htons(CONNECTIONSTATE_RESPONSE);
		    answer->head.totalsize =
			htons(HEADER_SIZE_10 +
			      sizeof(EIBNETIP_CONNECTIONSTATE_RESPONSE));
		    connectionstate_response =
			(EIBNETIP_CONNECTIONSTATE_RESPONSE *) ((byte *)
							       answer +
							       HEADER_SIZE_10);
		    connectionstate_response->channelid =
			received_channelid;

		    /* if the request has the wrong channelid set error code */
		    if (received_channelid != active_connection->channelid
			|| active_connection->reset)
			connectionstate_response->status = E_CONNECTION_ID;
		    else {	/* no error */

			connectionstate_response->status = E_NO_ERROR;
			active_connection->lastaccess =
			    time((time_t *) NULL);
		    }

		    sendPacket(answer,
			       make_Sockaddr((byte *) control_endpoint));
		    alarm(HEARTBEAT_REQUEST_TIME_OUT);	/* reset alarm */

		    free(answer);

#ifdef DEBUG
		    fprintf(stdout, "DEBUG: heartbeat answer sent \n");
#endif

		}

	    }
	    break;

	case (DISCONNECT_REQUEST):
	    {
#ifdef DEBUG
	      fprintf(stdout,"DEBUG: Disconnect request received \n");
	      zeig(received,htons(received_packet->head.totalsize));
#endif
		/* if the packet has not the right length discard it */
		if (htons(received_packet->head.totalsize) ==
		    sizeof(EIBNETIP_HPAI) + 2 + HEADER_SIZE_10) {
		    
		     /*
		     * *** extract disconnect request ***
		     */

		    EIBNETIP_HPAI *hpai =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10 + 2);
		    byte received_channel =
			*((byte *) received_packet + HEADER_SIZE_10);

		    /*
		     * *** prepare answer ***
		     */

		    EIBNETIP_PACKET *answer = malloc(HEADER_SIZE_10 + 2);
		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype = htons(DISCONNECT_RESPONSE);
		    answer->head.totalsize =
			htons(HEADER_SIZE_10 +
			      sizeof(EIBNETIP_DISCONNECT_RESPONSE));

		    disconnect_response =
			(EIBNETIP_DISCONNECT_RESPONSE *) ((byte *) answer +
							  HEADER_SIZE_10);
		    disconnect_response->channelid = received_channel;

		    /*if it is the wrong channel set connection id error flag */
		    if (received_channel != active_connection->channelid
			|| active_connection->reset)
			disconnect_response->status = E_CONNECTION_ID;
		    else {	/* disconnect client */

			alarm(0);	/* cancel alarm */
			sem_wait(semid);	/* is shared memory free ? */
			active_connection->reset = 1;	/* reset connection */
			sem_post(semid);	/* free shared memory */

			disconnect_response->status = E_NO_ERROR;	/* set the no error flag */

#ifdef DEBUG
			fprintf(stdout, "DEBUG: client disconnected \n");
#endif
		    }

		    /* send packet to the client */
		    sendPacket(answer, make_Sockaddr((byte *) hpai));

		    free(answer);
		}

		break;
	    }

	case (TUNNELLING_REQUEST):
	    {
	      
		EIBNETIP_COMMON_CONNECTION_HEADER *cch =
		    (EIBNETIP_COMMON_CONNECTION_HEADER *) ((byte *)
							   received_packet
							   +
							   HEADER_SIZE_10);
#ifdef DEBUG
		fprintf(stdout, "DEBUG: tunnelling request received\n");
		      zeig(received,htons(received_packet->head.totalsize));
#endif


		/* is the right channelid ? */
		if (cch->channelid == active_connection->channelid) {
		    /*
		     * *** prepare tunnelling ack ***
		     */

		    EIBNETIP_PACKET *answer =
			malloc(HEADER_SIZE_10 +
			       sizeof(EIBNETIP_COMMON_CONNECTION_HEADER));
		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype = htons(TUNNELLING_ACK);
		    answer->head.totalsize =
			htons(HEADER_SIZE_10 +
			      sizeof(EIBNETIP_COMMON_CONNECTION_HEADER));

		    /* prepare common connection header */
		    head = (EIBNETIP_COMMON_CONNECTION_HEADER *) ((byte *)
								  answer +
								  HEADER_SIZE_10);
		    head->structlength = 4;
		    head->channelid = active_connection->channelid;
		    head->sequencecounter =
			active_connection->sequenceReceive;

		    /* if the sequence is the expected */

		    if (cch->sequencecounter ==
			active_connection->sequenceReceive
			&& active_connection->reset == 0) {
			CEMI_L_DATA_MESSAGE *cemi =
			    (CEMI_L_DATA_MESSAGE *) ((byte *)
						     received_packet +
						     HEADER_SIZE_10 +
						     sizeof
						     (EIBNETIP_COMMON_CONNECTION_HEADER));

			/* we accept only L_DATA messages */
			if (cemi->mc == 0x11 || cemi->mc == 0x2E
			    || cemi->mc == 0x29) {

			    if (write(send_eib_pipe,	/*write message to pipe */
				      cemi,
				      sizeof(CEMI_L_DATA_MESSAGE) +
				      cemi->apci_length) < 0) {
				bailOut("Error: can't write to sendpipe");
			    }

			}
#ifdef DEBUG
			else {
			    fprintf(stdout,
				    "DEBUG: cEMI messagetype not supported\n");
			}
#endif

			sem_wait(semid);
			active_connection->sequenceReceive++;
			sem_post(semid);

		    }

		    if (cch->sequencecounter ==
			active_connection->sequenceReceive - 1
			|| cch->sequencecounter ==
			active_connection->sequenceReceive)
			head->typespecific = E_NO_ERROR;	/* return no error flag */
		    else
			head->typespecific = E_SEQUENCE_NUMBER;	/* return e sequencenumber error */

		    sendPacket(answer, active_connection->data_endpoint);	/* send packet to client */

		    free(answer);
		}

		break;
	    }

	case (TUNNELLING_ACK):
	    {

		EIBNETIP_COMMON_CONNECTION_HEADER *cch =
		    (EIBNETIP_COMMON_CONNECTION_HEADER *) ((byte *)
							   received_packet
							   +
							   HEADER_SIZE_10);


#ifdef DEBUG
		fprintf(stdout, "DEBUG: tunneling ack received\n");
		zeig(received,htons(received_packet->head.totalsize));
#endif

		if (cch->channelid == active_connection->channelid
		    && cch->sequencecounter ==
		    active_connection->sequenceSend
		    && active_connection->reset == 0) {
		    sem_wait(semid);
		    active_connection->sequenceSend++;	/* increment sequencecounter */
		    sem_post(semid);

		    if (kill(pid_SendUdp, SIGUSR1) == -1) {	/* send signal to Send UDP */
			bailOut("Error: can't send signal to send_udp");
		    }

		}
		break;

	    }
	case (DISCONNECT_RESPONSE):
	    {
		/*simple don't care: it don't matter if we get it or not */
		break;
	    }

	case (SEARCH_REQUEST):
	    {

#ifdef DEBUG
		fprintf(stdout, "DEBUG: Search request received\n");
		zeig(received,htons(received_packet->head.totalsize));
#endif

		if (htons(received_packet->head.totalsize) ==
		    HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI)) {

		    EIBNETIP_HPAI *hpai =
			(EIBNETIP_HPAI *) ((byte *) received_packet +
					   HEADER_SIZE_10);

		    /*
		     * *** prepare search response ***
		     */

		    EIBNETIP_PACKET *answer = (EIBNETIP_PACKET *) malloc(HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI) + sizeof(EIBNETIP_DEVINF_DIB) + 6);	/*6 is sizeof eibnet/Ip suppfam dib */
		    answer->head.headersize = HEADER_SIZE_10;
		    answer->head.version = EIBNETIP_VERSION_10;
		    answer->head.servicetype = htons(SEARCH_RESPONSE);
		    answer->head.totalsize =
			htons(HEADER_SIZE_10 + sizeof(EIBNETIP_HPAI) +
			      sizeof(EIBNETIP_DEVINF_DIB) +
			      sizeof(EIBNETIP_SUPPFAM_DIB));

		    /* make Server Hpai at the specified location */
		    makeServerHpai((byte *) answer + HEADER_SIZE_10, addr);

		    /* make device information at the specified location */
		    make_DEV_INF_DIB((byte *) answer + HEADER_SIZE_10 +
				     sizeof(EIBNETIP_HPAI));

		    /* make supported families descripiton */
		    make_SUPPFAMDIB((byte *) answer + HEADER_SIZE_10 +
				    sizeof(EIBNETIP_HPAI) +
				    sizeof(EIBNETIP_DEVINF_DIB));

		    sendPacket(answer, make_Sockaddr((byte *) hpai));

		}

		break;

	    }
	default:
	    {

#ifdef DEBUG
		fprintf(stdout, "DEBUG: Unknown servicetype: %X \n",
			htons(received_packet->head.servicetype));
#endif
		break;
	    }
	}

    }

}


/*
 * ***************************************************************** EOF ***
 */
